#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to create a drawing look using G'mic Engrave
#      Inspired by a Nidhogg's tutorial posted on Gimpchat

# (c) Dinasset 2020
#
#   History:
#
#   v1.0: 2020-11-09: Initial version
#   v1.1: 2020-11-11: modified Contours
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import time
import datetime
from gimpfu import *
from array import array

# define here some constant value, when required
wrkwh=2400  #3000
selgauss=5
do_white_balance=True
blacks_to_remove=24   
whites_to_remove=12
poster_perc=40   
poster_levels=8

############################################################################
# Service routines

#---------------------------------------------------------------------------
# for messages: include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    gimp.message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    gimp.message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

#---------------------------------------------------------------------------
# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inDrawable, widthheight):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inDrawable.name
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    if widthheight == 0 :   # means the user wants to have output size same as input size
        widthheight = realWidthHeight
    optiWidthHeight = widthheight       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size...  
   # copy the source layer...
   # scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    
    isamask = pdb.gimp_item_is_layer_mask (inDrawable)
    if isamask==True:    
        inLayer = pdb.gimp_layer_from_mask (inDrawable)
        newBG = pdb.gimp_layer_new_from_drawable (inLayer, newImage)
    else:    
        newBG = pdb.gimp_layer_new_from_drawable (inDrawable, newImage)
          
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewBG from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    
    # keep a copy of the original background below the one which will be used  
    #
    Original = newBG.copy()
    newImage.add_layer (Original, 1)
    Original.name = oname+" (original)"
    pdb.gimp_item_set_visible (Original, False)
   
   # do a selective gaussian blur (added)
    if selgauss>0:
        pdb.plug_in_sel_gauss (newImage, newBG, selgauss*1.0, 50)
    
   # do white balancing to colours values ------------
    if do_white_balance==True:  
        newBGwb=WhiteBalanceStretch (newImage, newBG)
        newBGwb.name = "NewBG after white balancing"
    # add saturation
    pdb.gimp_drawable_hue_saturation (newBGwb, HUE_RANGE_ALL, 0., 0., 33.0, 0.0)
    newBGwb.opacity = 100.0
    
    return(newImage)
#--------------------------------------------------------------------------- 
def DeCast (inImage, inDrawable, castR, castG, castB, howmuch) :    
    
    # Reduces the cast in a layer, using the 3 calculated percentages of cast.
    # Quite fast, since it makes use of regions.
    # Thanks to Akkana Peck
        
    # Indicates that the process has started.
    gimp.progress_init("De-casting " + inDrawable.name + "...")

    layername = "Modified: " + inDrawable.name
    # size for the new layer
    newWidth = inDrawable.width           
    newHeight = inDrawable.height         
        
    # Create the new layer:
    destDrawable = gimp.Layer(inImage, layername, newWidth, newHeight,
                              inDrawable.type, inDrawable.opacity, inDrawable.mode)
    inImage.add_layer(destDrawable, 0)

    # No need to clear the layer here --
    # we'll initialize it to zero when we create the pixel array.

    srcWidth, srcHeight = inDrawable.width, inDrawable.height
    srcRgn = inDrawable.get_pixel_rgn(0, 0, srcWidth, srcHeight,
                                     False, False)
    src_pixels = array("B", srcRgn[0:srcWidth, 0:srcHeight])

    dstRgn = destDrawable.get_pixel_rgn(0, 0, newWidth, newHeight,
                                            True, True)
        
    p_size = len(srcRgn[0,0])                  # pixel size is 3 !
        
    dest_pixels = array("B", "\x00" * (newWidth * newHeight * p_size))   # initialized to 0

    # Loop over the region:                    
    for x in xrange(0, srcWidth - 1) :
        for y in xrange(0, srcHeight) :
            newx = x        
            newy = y        
                        
            src_pos = (x + srcWidth * y) * p_size
            dest_pos = (newx + newWidth * newy) * p_size
                        
            srcval = src_pixels[src_pos: src_pos + p_size]
            newval = src_pixels[src_pos: src_pos + p_size]
                        
            #  
            newval[0] = min(int(newval[0]/castR),255)
            newval[1] = min(int(newval[1]/castG),255)
            newval[2] = min(int(newval[2]/castB),255)
            #  now balance with existing values using 'howmuch' -------------- 
            newval[0] = int((newval[0]*howmuch)+(srcval[0]*(1.0-howmuch)))
            newval[1] = int((newval[1]*howmuch)+(srcval[1]*(1.0-howmuch)))
            newval[2] = int((newval[2]*howmuch)+(srcval[2]*(1.0-howmuch)))
            # ----------------------------------------------------------------
            #
            dest_pixels[dest_pos : dest_pos + p_size] = newval
                        
        # Docs say progress_update takes a percent,
        # but it's really a fraction.
        progress = float(x)/inDrawable.width
        if (int(progress * 100) % 20 == 0) :
            gimp.progress_update(progress)
                        
    # Copy the whole array back to the pixel region:
    dstRgn[0:newWidth, 0:newHeight] = dest_pixels.tostring() 

    destDrawable.flush()
    destDrawable.merge_shadow(True)
    destDrawable.update(0, 0, newWidth,newHeight)

    # Remove the old layer
    inImage.remove_layer(inDrawable)

    # End progress.
    pdb.gimp_progress_end()
 
    return(destDrawable)
    
#--------------------------------------------------------------------------- 
# apply a white balancing to the source image
#---------------------------------------------------------------------------
def WhiteBalanceStretch (inImage, inDrawable) :

    theImage = inImage
    theDrawable = inDrawable
    
    # remember current values distribution 0-0.5, 0.5-1.0
    # 
    (mean, S, M, Pix, countDark, Perc) = pdb.gimp_drawable_histogram (theDrawable,
                                                                      HISTOGRAM_VALUE, 0, 0.5)   
    (mean, S, M, Pix, countBright, Perc) = pdb.gimp_drawable_histogram (theDrawable,
                                                                        HISTOGRAM_VALUE, 0.5, 1.0)   
    sourceratio = ((countDark*1.0)+1.0)/((countBright*1.0)+1.0)
    
    Processed = theDrawable.copy()
    Processed.name = "Processed Layer"
    theImage.add_layer(Processed, 0)
     
    # decast first, not to have final outcome with "empties extremes"
    # Calculate the cast
    (meanR, S, M, Pix, count, percent) = pdb.gimp_drawable_histogram (Processed,
                                                                      HISTOGRAM_RED, 0., 1.)        
    (meanG, S, M, Pix, count, percent) = pdb.gimp_drawable_histogram (Processed,
                                                                      HISTOGRAM_GREEN, 0., 1.)     
    (meanB, S, M, Pix, count, percent) = pdb.gimp_drawable_histogram (Processed,
                                                                      HISTOGRAM_BLUE, 0., 1.)      
    meangray = int((meanR+meanG+meanB)/3)        # calculate the average
    castR = float((meanR*1.000)/(meangray*1.000))
    castG = float((meanG*1.000)/(meangray*1.000))
    castB = float((meanB*1.000)/(meangray*1.000))
    howmuch = 0.33  # how much apply decasting
    Processed = DeCast (theImage, Processed, castR, castG, castB, howmuch)
    Processed.name = "Processed Layer de-casted"
    
    # ===============================================
    # COMPRESS COLOURS REMOVING EXTREME VALUES
    # ===============================================
    
   # suppress extreme black and extreme white pixels ------------------
   # --- hence apply the Gimp levels such that the drawable
   # --- will have values starting from blacks_to_remove to 255-whites_to_remove
    level_start_value = 0.0; level_end_value = 1.0  # defaults
    if int(blacks_to_remove)>0:
        level_start_value = float(blacks_to_remove*1.0/255.0)
    if int(whites_to_remove)>0:
        level_end_value = 1.00-float(whites_to_remove*1.0/255.0)
    if int(blacks_to_remove)>0 or int(whites_to_remove)>0:    
        pdb.gimp_drawable_levels (Processed, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.00, level_start_value, level_end_value, False)  
    
    return(Processed)

############################################################################
# Main
#---------------------------------------------------------------------------

def DrawStyleJ (inImage, inLayer,
                widthheight, Mask,
                doFlatten):
    
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    if widthheight == 0 :   # means the user wants to have output size same as input size
        widthheight = inImage.width+inImage.height      # real size of the input image
     
    newImage = createOutimage (inImage, inLayer, wrkwh)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    Original = newImage.layers[2]
    Preprocessed = newImage.layers[0]
    PpW = Preprocessed.width
    PpH = Preprocessed.height
    PpWH = Preprocessed.width+Preprocessed.height
    # add grain
    GrainTexture = pdb.gimp_layer_new (newImage, PpW, PpH, RGBA_IMAGE, "Texture",
                                       100.0, LAYER_MODE_GRAIN_MERGE)
    newImage.add_layer (GrainTexture, 0)
    pdb.gimp_context_set_pattern ("3D Green")
    pdb.gimp_drawable_edit_fill (GrainTexture, FILL_PATTERN)
    pdb.gimp_drawable_desaturate (GrainTexture, DESATURATE_LUMA)
    pdb.plug_in_spread (newImage, GrainTexture, 5.0, 5.0)
    Preprocessed = pdb.gimp_image_merge_down (newImage, GrainTexture, CLIP_TO_IMAGE)
    
    # check mean
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (Preprocessed,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    
    mofified=False
    if meanl < 110:
        #info_msg("mean luminance = "+str(meanl))
        while meanl < 110:
            pdb.gimp_drawable_levels (Preprocessed, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.05, 0.01, 1.0, False)
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (Preprocessed,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        mofified=True
        #info_msg("modified mean luminance = "+str(meanl))
    
    #pdb.gimp_display_new (newImage)
    #return
        
    # step: G'MIC Simple Local Contrast
    # [G'MIC] Simple Local Contrast: simplelocalcontrast_p 25,2,50,1,1,1,1,1,1,1,1,1
    pdb.plug_in_gmic_qt(newImage, Preprocessed, 1,2 ,
        "-v - -simplelocalcontrast_p 25,2,50,1,1,1,1,1,1,1,1,1 ")  #last 1=no parallel process
    Contrasted = newImage.layers[0]
    Contrasted.name = "- Contrasted"
    
    #pdb.gimp_display_new (newImage)
    #return
    
    #-------------------------------------------------------------- 
    # step: G'MIC Engrave
    # to engrave the whole image it's useful move values from 0-255 to 0-128
    BaseEngrave = Contrasted.copy()
    newImage.add_layer (BaseEngrave, 0)
    BaseEngrave.name = "- adj for input to Engrave"
    pdb.gimp_drawable_levels (BaseEngrave, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.00, 0.1, 0.5, False)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # [G'MIC] Engrave: fx_engrave 0.5,50,0,8,40,0,0,0,10,1,0,0,0,1,0,50,50    < std dft
    pdb.plug_in_gmic_qt(newImage, BaseEngrave, 1,2 ,
        "-v - -fx_engrave 0.5,50,0,8,40,0,0,"+
        "0"+                                    # 0=only black lines, 1=w/coloured BG
        ",10,1,0,0,0,1,0,50,50 ")
    
    Engraved = newImage.layers[0]
    Engraved.name = "- lines of Engrave"
    Engraved.mode = LAYER_MODE_LINEAR_BURN      #MULTIPLY    #DARKEN_ONLY
    Engraved.opacity = 85.0
    pdb.gimp_drawable_levels (Engraved, HISTOGRAM_VALUE,
                              0.0, 1.0, False, 1.0, 0.4, 0.9, False)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # build the BG for the engraving lines
    EngraveBGposter = Preprocessed.copy()
    newImage.add_layer(EngraveBGposter, 1)   # below EngraveBGasis (see next def)
    EngraveBGposter.name = "- BG of G'mic Engrave (posterized)"
    pdb.gimp_drawable_posterize (EngraveBGposter, poster_levels)
    pdb.plug_in_gauss (newImage, EngraveBGposter, 3.0, 3.0, 0)          
    pdb.gimp_drawable_curves_spline (EngraveBGposter, HISTOGRAM_VALUE, 4, (0.0,0.1, 1.0,1.0))
    EngraveBGasis = Preprocessed.copy()
    newImage.add_layer(EngraveBGasis, 1)     # below Engraved
    EngraveBGasis.name = "- BG of G'mic Engrave (as is)"
    EngraveBGasis.opacity = (100.0-poster_perc*1.0)  # = 0 if poster_perc at 100%
    pdb.gimp_drawable_curves_spline (EngraveBGasis, HISTOGRAM_VALUE, 4, (0.0,0.1, 1.0,1.0))
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # add contours to BG poster ---------------------------------------
    ContoursBG = EngraveBGposter.copy()
    newImage.add_layer (ContoursBG, 1)
    ContoursBG.name = "Contours"
    
    #pdb.gimp_display_new (newImage)
    #return
    
    pdb.plug_in_edge (newImage, ContoursBG, 1.2, 0, 2)          # chg 11.11 rel.1.01
    # edge type = {SOBEL(0),PREWITT(1),GRADIENT(2),ROBERTS(3),DIFFERENTIAL(4),LAPLACE(5) }
    pdb.gimp_drawable_curves_spline (ContoursBG, HISTOGRAM_VALUE, 4, (0.0,0.1, 1.0,1.0))
    pdb.gimp_drawable_desaturate (ContoursBG, DESATURATE_LUMA)   
    ContoursBG.mode = LAYER_MODE_SOFTLIGHT  #LUMA_LIGHTEN_ONLY  # chg 11.11 rel 1.01
    ContoursBG.opacity = 50.0    
    
    #pdb.gimp_display_new (newImage)
    #return
     
    ContoursTOP = ContoursBG.copy()
    newImage.add_layer (ContoursTOP, 0)
    pdb.plug_in_vinvert (newImage, ContoursTOP)
    ContoursTOP.name = "Contours on Top inverted for Chroma"
    ContoursTOP.mode = LAYER_MODE_OVERLAY  #SOFTLIGHT #LCH_CHROMA
    ContoursTOP.opacity = 100.0
    pdb.plug_in_gauss (newImage, ContoursTOP, PpW/20, PpH/20, 0) 
    
    #-------------------------------------------------------------- 
    # final steps
    
    if Mask>0:
        maskedBG = Original.copy()
        newImage.add_layer (maskedBG, 0)
        pdb.gimp_item_set_visible (maskedBG, True)
        maskedBG.mode = LAYER_MODE_NORMAL                                 
        maskedBG.opacity = 50.0                                             
        #                                                                     
        if Mask==1:
            blackmask = pdb.gimp_layer_create_mask (maskedBG, ADD_MASK_BLACK)    
            pdb.gimp_layer_add_mask (maskedBG, blackmask)                        
        if Mask==2:
            graymask = pdb.gimp_layer_create_mask (maskedBG, ADD_MASK_COPY)      
            pdb.gimp_layer_add_mask (maskedBG, graymask)                         
         
    uratio = widthheight*1.0/PpWH*1.0
    uwidth = int(PpW*uratio); uheight = int(PpH*uratio)
    pdb.gimp_image_scale (newImage, uwidth, uheight)   # to user desired size
    
    if doFlatten==True:
        if Mask>0:
            pdb.gimp_item_set_visible (maskedBG, False)
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        if Mask>0:
            pdb.gimp_item_set_visible (maskedBG, True)
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" as an engraved drawing"
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="Nidhogg/dinasset"
menu="<Image>/Diego/Drawing/"
copyrightYear="2020"
desc="Draw Effect Style J (v.1.1)"

register(
    "DrawStyleJ",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+"...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 7200, 100)),
        (PF_OPTION, "maskbg", "Add Masked BG on Top?"+
                    "\n  (for final user adjustments)", 0, ["*none",
                                                            "black",
                                                            "gray copy"]),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    DrawStyleJ,
    menu=menu
)

main()
